from typing import Any, Mapping, Optional, Union

from jinja2 import Environment, Template
from jinja2.environment import TemplateModule
from psycopg.sql import SQL, Composed

from .extension import PsycopgExtension
from .context import FormatArgsContext
from .sql import IntoSql, sql_filter, sql_join_filter

CONTEXT = FormatArgsContext("format_args")
_NO_VALUE = object()


def psycopg_filter(value: Any):
    """Jinja filter that saves the value inside a dictionary in ContextVar
    and returns a psycopg format placeholder such as `{key}`"""
    if isinstance(value, IntoSql):
        value = value.__sql__()

    if isinstance(value, SQL):
        # No need to pass SQL to psycopg's formatter,
        # since it's included as is
        return value.as_string(None)

    key = CONTEXT.save_value(value)
    return f"{{{key}}}"


def escape_percents(composed: Composed) -> Composed:
    """Replace all occurences of '%' with '%%',
    so that psycopg doesn't mistake them for placeholders.
    Actual placeholders can still be created with psycopg.sql.Placeholder"""
    new_sequence = []
    for token in composed:
        if isinstance(token, SQL):
            token = SQL(token.as_string(None).replace("%", "%%"))
        new_sequence.append(token)

    return Composed(new_sequence)


class SqlTemplate:
    """Wrapper for jinja2.Template that stores static format arguments
    such as `{{ 'text' }}`"""

    def __init__(self, template: Template, static_args: dict[str, Any]) -> None:
        self._template = template
        self._static_args = static_args

    def render(self, *args, **kwargs) -> Composed:
        """Same as jinja2.Template.render, but returns a psycopg.sql.Composed object"""
        recorder = CONTEXT.recorder("dynamic")
        with recorder:
            sql = SQL(self._template.render(*args, **kwargs))
        dynamic_args = recorder.unwrap()

        composed = sql.format(**self._static_args, **dynamic_args)
        return escape_percents(composed)

    def make_module(
        self,
        vars: Optional[dict[str, Any]] = None,
        shared: bool = False,
        locals: Optional[Mapping[str, Any]] = None,
    ):
        """Same as jinja2.Template.make_module, but returns a wrapper that remembers all the format arguments"""
        recorder = CONTEXT.recorder("dynamic")
        with recorder:
            module = self._template.make_module(vars, shared, locals)
        dynamic_args = recorder.unwrap()

        return SqlTemplateModule(module, {**self._static_args, **dynamic_args})


class SqlTemplateModule:
    """Wrapper over jinja2.environment.TemplateModule that stores all the format arguments
    for use in SQL.format"""

    def __init__(self, module: TemplateModule, args: dict[str, Any]) -> None:
        self._module = module
        self._args = args

    def render(self):
        """Returns a formatted SQL statement"""
        composed = SQL(str(self._module)).format(**self._args)
        return escape_percents(composed)

    @property
    def inner(self):
        return self._module

    def getattr(self, name: str, default: Any = _NO_VALUE) -> Any:
        """Get attribute of the inner module"""
        if default is _NO_VALUE:
            return getattr(self._module, name)
        else:
            return getattr(self._module, name, default)


class JinjaPsycopg:
    """Wrapper over jinja2.Environment that generates `SqlTemplate`s"""

    def __init__(self, env: Optional[Environment] = None) -> None:
        self._env = env or Environment()
        self._prepare_environment()

    def _prepare_environment(self):
        self._env.add_extension(PsycopgExtension)
        self._env.filters["psycopg"] = psycopg_filter
        self._env.filters["sql"] = sql_filter
        self._env.filters["sqljoin"] = sql_join_filter

    def from_string(self, source: str) -> SqlTemplate:
        # Jinja2 processes its blocks in two iterations:
        # static values like {{ 'text' }} are processed during from_string,
        # and dynamic ones during Template.render or make_module
        recorder = CONTEXT.recorder("static")
        with recorder:
            template = self._env.from_string(source)

        return SqlTemplate(template, recorder.unwrap())

    def render(
        self, template: Union[str, SqlTemplate], params: dict[str, Any] = {}
    ) -> Composed:
        if isinstance(template, str):
            template = self.from_string(template)

        return template.render(params)



from dataclasses import dataclass
from textwrap import dedent
import pytest
import psycopg
from psycopg import Connection, sql

from jinja_psycopg import JinjaPsycopg


@pytest.fixture
def conn():
    connection = psycopg.connect()
    yield connection
    connection.close()


def test_args(conn: Connection):
    query = "SELECT * FROM {{ table }} WHERE id = {{ id }}"
    expected = 'SELECT * FROM "sources" WHERE id = 5'
    params = {"table": sql.Identifier("sources"), "id": 5}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


def test_no_args(conn: Connection):
    query = "SELECT * FROM aaa"

    assert JinjaPsycopg().render(query).as_string(conn) == query


def test_string(conn: Connection):
    query = "WHERE name = {{ 'foo' }}"
    expected = "WHERE name = 'foo'"

    assert JinjaPsycopg().render(query).as_string(conn) == expected


def test_string_order(conn: Connection):
    query = "WHERE name = {{ foo }}, value = {{ 'bar' }}"
    expected = "WHERE name = 'foo', value = 'bar'"
    params = {"foo": "foo"}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


class CustomRenderer(JinjaPsycopg):
    def _prepare_environment(self):
        super()._prepare_environment()

        def appendA(string: str):
            return string + "A"

        def appendB(string: str):
            return string + "B"

        self._env.globals["appendA"] = appendA
        self._env.filters["appendB"] = appendB


def test_function(conn: Connection):
    query = "VALUES ( {{ appendA('bar') }} )"
    expected = "VALUES ( 'barA' )"

    assert CustomRenderer().render(query).as_string(conn) == expected


def test_function_order(conn: Connection):
    query = "VALUES ( {{ foo }}, {{ appendA('bar') }} )"
    expected = "VALUES ( 'foo', 'barA' )"
    params = {"foo": "foo"}

    assert CustomRenderer().render(query, params).as_string(conn) == expected


def test_filter(conn: Connection):
    query = "VALUES ( {{ 'bar' | appendB }} )"
    expected = "VALUES ( 'barB' )"

    assert CustomRenderer().render(query).as_string(conn) == expected


def test_filter_order(conn: Connection):
    query = "VALUES ( {{ foo }}, {{ 'bar' | appendB }} )"
    expected = "VALUES ( 'foo', 'barB' )"
    params = {"foo": "foo"}

    assert CustomRenderer().render(query, params).as_string(conn) == expected


def test_manual_psycopg(conn: Connection):
    query = "SELECT * FROM {{ table | psycopg }}"
    expected = 'SELECT * FROM "jsons"'
    params = {"table": sql.Identifier("jsons")}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


@pytest.mark.parametrize(
    "flag,expected",
    [
        (False, "SELECT * FROM sources"),
        (True, "SELECT * FROM sources WHERE id = 5"),
    ],
)
def test_if(conn: Connection, flag: bool, expected: str):
    query = "SELECT * FROM sources{% if flag %} WHERE id = {{ id }}{% endif %}"
    params = {"flag": flag, "id": 5}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


def test_for(conn: Connection):
    query = dedent(
        """\
        {% for column in columns -%}
        ADD COLUMN {{ column }} TEXT{% if not loop.last %},
        {% endif %}{% endfor %}\
        """
    )
    expected = dedent(
        """\
        ADD COLUMN "id" TEXT,
        ADD COLUMN "source" TEXT,
        ADD COLUMN "html" TEXT\
        """
    )
    params = {
        "columns": [
            sql.Identifier("id"),
            sql.Identifier("source"),
            sql.Identifier("html"),
        ]
    }

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


def test_for_static(conn: Connection):
    query = dedent(
        """\
        {% for ident in idents -%}
        {{ 'text' }} {{ ident }}{% if not loop.last %},
        {% endif %}{% endfor %}\
        """
    )
    expected = dedent(
        """\
        'text' "foo",
        'text' "bar",
        'text' "fluff"\
        """
    )
    params = {
        "idents": [
            sql.Identifier("foo"),
            sql.Identifier("bar"),
            sql.Identifier("fluff"),
        ]
    }

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


def test_module(conn: Connection):
    query = dedent(
        """\
        {% set val = 1 -%}
        SELECT * FROM {{ table }}"""
    )
    expected = 'SELECT * FROM "sources"'
    params = {"table": sql.Identifier("sources")}

    module = JinjaPsycopg().from_string(query).make_module(params)
    rendered = module.render().as_string(conn)

    assert rendered == expected
    assert getattr(module.inner, "val") == 1


def test_module_getattr():
    query = "{% set val = 1 %}"

    module = JinjaPsycopg().from_string(query).make_module()
    assert module.getattr("val") == 1


def test_module_default():
    query = "{% set val = 1 %}"

    module = JinjaPsycopg().from_string(query).make_module()
    assert module.getattr("foo", 5) == 5


def test_module_getattr_error():
    query = "{% set val = 1 %}"

    module = JinjaPsycopg().from_string(query).make_module()
    with pytest.raises(AttributeError):
        module.getattr("foo")


def test_module_static(conn: Connection):
    query = dedent(
        """\
        {% set val = 1 -%}
        {{ 'text' }} {{ table }}"""
    )
    expected = "'text' \"sources\""
    params = {"table": sql.Identifier("sources")}

    sql_module = JinjaPsycopg().from_string(query).make_module(params)
    rendered = sql_module.render().as_string(conn)

    assert rendered == expected
    assert sql_module._module.val == 1  # type:ignore


def test_sql_filter(conn: Connection):
    query = "SELECT * FROM {{ 'foo' | sql }}"
    expected = "SELECT * FROM foo"

    assert JinjaPsycopg().render(query).as_string(conn) == expected


@dataclass
class Table:
    schema: str
    name: str

    def __sql__(self):
        return sql.SQL("{}.{}").format(
            sql.Identifier(self.schema), sql.Identifier(self.name)
        )


def test_into_sql(conn: Connection):
    query = "SELECT * FROM {{ table }}"
    expected = 'SELECT * FROM "public"."items"'
    params = {"table": Table("public", "items")}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


def test_join(conn: Connection):
    query = "{{ values | sqljoin(',\n') }}"
    expected = dedent(
        """\
        "foo",
        "sources"."bar\""""
    )
    params = {"values": [sql.Identifier("foo"), Table("sources", "bar")]}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


@dataclass
class Column:
    name: str
    type: str

    @property
    def add_column(self):
        return sql.SQL("ADD COLUMN {} {}").format(
            sql.Identifier(self.name), sql.SQL(self.type)
        )


def test_join_attribute(conn: Connection):
    query = "{{ columns | sqljoin(',\n', attribute='add_column') }}"
    expected = dedent(
        """\
        ADD COLUMN "id" SERIAL PRIMARY KEY,
        ADD COLUMN "name" TEXT"""
    )
    params = {"columns": [Column("id", "SERIAL PRIMARY KEY"), Column("name", "TEXT")]}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected


def test_escape_percent(conn: Connection):
    query = dedent(
        """\
        SELECT * FROM foo
        WHERE field1 LIKE '%'
        AND field2 = {{ field2 }}"""
    )
    expected = dedent(
        """\
        SELECT * FROM foo
        WHERE field1 LIKE '%%'
        AND field2 = %(field2)s"""
    )
    params = {"field2": sql.Placeholder("field2")}

    assert JinjaPsycopg().render(query, params).as_string(conn) == expected



from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import Optional, Any


@dataclass
class FormatArgs:
    """Data structure for recording values in jinja blocks to be formatted by psycopg
    Args:
        - prefix (str): Prefix used in dictionary keys"""

    prefix: str
    dictionary: dict[str, Any] = field(default_factory=dict)
    num_values: int = 0

    def save_value(self, value: Any) -> str:
        """Saves the value to dictionary, returning its key
        in the format of `prefix#number`"""
        key = f"{self.prefix}#{self.num_values}"
        self.dictionary[key] = value

        self.num_values += 1
        return key


class FormatArgsContext:
    """Wrapper for ContextVar used for saving format args from within `psycopg` filter
    and for creating argument recorders"""

    def __init__(self, name: str) -> None:
        """Args:
        - name (str): name used by the ContextVar"""
        self._context_var = ContextVar[Optional[FormatArgs]](name, default=None)

    def save_value(self, value: Any) -> str:
        """Saves the value to dictionary, returning its key
        in the format of `prefix#number`.

        Errors if ContextVar is empty"""
        context = self._context_var.get()
        if context is None:
            raise RuntimeError(
                "Called ContextWriter.save_value, but no context was found"
            )

        return context.save_value(value)

    def recorder(self, prefix: str):
        """Args:
        - prefix (str): Prefix for the keys in the resulting dictionary"""
        return FormatArgsRecorder(self._context_var, prefix)


class FormatArgsRecorder:
    """ContextVar wrapper that works as a context manager
    and records arguments saved within its scope into a dictionary"""

    def __init__(
        self, context_var: ContextVar[Optional[FormatArgs]], prefix: str
    ) -> None:
        self._context_var = context_var
        self._prefix = prefix
        self._recorded = None

    def __enter__(self):
        self._token = self._context_var.set(FormatArgs(self._prefix))

    def __exit__(self, type, value, traceback):
        context = self._context_var.get()
        if context is None:
            raise RuntimeError("Finished recording, but ContextVar was empty")

        self._recorded = context.dictionary
        self._context_var.reset(self._token)

    def unwrap(self) -> dict[str, Any]:
        """Returs the recorded arguments, errors if nothing was recorder"""
        if self._recorded is None:
            raise RuntimeError(
                "Called ContextDictRecorder.unwrap(), but nothing was recorded"
            )

        return self._recorded



from typing import Optional, Iterable, Protocol, Union, runtime_checkable
from psycopg.sql import SQL, Composable


@runtime_checkable
class IntoSql(Protocol):
    def __sql__(self) -> Composable:
        ...


SqlLike = Union[Composable, IntoSql]


def sql_filter(value: str) -> SQL:
    """Jinja filter for including a sql string as is
    Usage: `{{ 'text' | sql }}`"""
    return SQL(value)


def _preprocess_before_join(items: Iterable[SqlLike], attribute: Optional[str] = None):
    for item in items:
        if attribute is not None:
            item = getattr(item, attribute)

        if isinstance(item, IntoSql):
            item = item.__sql__()

        yield item


def sql_join_filter(
    value: Iterable[SqlLike], delimiter: str, attribute: Optional[str] = None
):
    return SQL(delimiter).join(_preprocess_before_join(value, attribute))



from typing import Iterable
from jinja2.ext import Extension
from jinja2.lexer import TokenStream, Token


class PsycopgExtension(Extension):
    """Wraps all expressions with the psycopg filter,
    so
    `{{ variable | filter1 | filter2 }}`
    becomes
    `{{ (variable | filter1 | filter2) | psycopg }}`
    Inspired by [jinjasql](https://github.com/sripathikrishnan/jinjasql)
    """

    def filter_stream(self, stream: TokenStream) -> Iterable[Token]:
        token_id = 0
        while not stream.eos:
            token = next(stream)
            token_id += 1

            if token.test("variable_begin"):
                var_expr: list[Token] = []
                while not token.test("variable_end"):
                    var_expr.append(token)
                    token = next(stream)
                    token_id += 1

                variable_end = token

                last_token = var_expr[-1]
                lineno = last_token.lineno

                if not last_token.test("name:psycopg"):
                    var_expr.insert(1, Token(lineno, "lparen", "("))
                    var_expr.append(Token(lineno, "rparen", ")"))
                    var_expr.append(Token(lineno, "pipe", "|"))
                    var_expr.append(Token(lineno, "name", "psycopg"))

                var_expr.append(variable_end)
                yield from var_expr
            else:
                yield token
